"""Console script for gitlab_usermgmt."""
import argparse
import sys
import gitlab_usermgmt.gitlab_usermgmt
import logging

def main():
    """Console script for gitlab_usermgmt."""
    parser = argparse.ArgumentParser(
        description='The tool that helps you manage users in big GitLab groups and projects.',
        epilog='For more information go to: https://gitlab.com/ayeks/gitlab_usermgmt'
    )

    # check if we need gitlab auth
    if any(c in sys.argv for c in ('-a', '--apply_access', '-l', '--list_users')):
        token_required = True
    else:
        token_required = False

    # check if we need group id
    if any(c in sys.argv for c in ('-l', ' --list_users')):
        group_id_required = True
    else:
        group_id_required = False

    parser.add_argument('-l', '--list_users',
                        help='writes users.json with all user access',
                        dest='list_users',
                        required=False,
                        action="store_true"
                        )
    parser.add_argument('-g', '--group_id',
                        help='the source group id',
                        dest='group_id',
                        required=group_id_required,
                        type=int)
    parser.add_argument('-t', '--gitlab_token',
                        help='your gitlab api token',
                        dest='gitlab_token',
                        required=token_required,
                        type=str)
    parser.add_argument('-u', '--gitlab_url',
                        help='your gitlab api token',
                        dest='gitlab_url',
                        required=False,
                        default="https://gitlab.com",
                        type=str)
    parser.add_argument('-c', '--compare_access',
                        help='Compare access with to an access definition json file',
                        dest='compare_access',
                        required=False,
                        action="store_true")
    parser.add_argument('-r', '--remove_expire_dates',
                        help='Remove expire_at entries for managed access rights.',
                        dest='remove_expire_dates',
                        required=False,
                        action="store_true")
    parser.add_argument('-m', '--manage_admins_only',
                        help='Removes access only from maintainers and owners, just sets expire dates at others',
                        dest='manage_admins_only',
                        required=False,
                        action="store_true")
    parser.add_argument('-a', '--apply_access',
                        help='Apply the results of the compared access',
                        dest='apply_access',
                        required=False,
                        action="store_true")
    parser.add_argument('-v', '--verbose',
                        help='Print debug log info',
                        action="store_true")
    args = parser.parse_args()

    log_format = '%(asctime)s - %(levelname)s - %(name)s - %(message)s'
    if args.verbose:
        logging.basicConfig(format=log_format, level=logging.DEBUG)
    else:
        logging.basicConfig(format=log_format, level=logging.INFO)

    for arg, value in sorted(vars(args).items()):
        logging.debug("Argument %s: %r", arg, value)

    if args.list_users:
        logging.debug('list_users starting ...')
        gitlab_usermgmt.gitlab_usermgmt.write_user_access_list(
            args.group_id, args.gitlab_token, args.gitlab_url)
        logging.debug('list_users finished.')

    if args.compare_access:
        logging.debug('compare_access starting ...')
        gitlab_usermgmt.gitlab_usermgmt.write_user_compared_list(
            "users.json",
            "access_definition.json",
            "users_compared.json"
        )
        logging.debug('compare_access finished.')

    if args.remove_expire_dates:
        logging.debug('remove_expire_dates starting ...')
        users = gitlab_usermgmt.gitlab_usermgmt.read_json_to_dict("users_compared.json")
        users = gitlab_usermgmt.gitlab_usermgmt.remove_expire_dates(users)
        gitlab_usermgmt.gitlab_usermgmt.write_dict_to_json(users, "users_compared.json")
        logging.debug('remove_expire_dates finished.')

    if args.manage_admins_only:
        logging.debug('manage_admins_only starting ...')
        users = gitlab_usermgmt.gitlab_usermgmt.read_json_to_dict("users_compared.json")
        gitlab_usermgmt.gitlab_usermgmt.manage_admins_only(users)
        gitlab_usermgmt.gitlab_usermgmt.write_dict_to_json(users, "users_compared.json")
        logging.debug('manage_admins_only finished.')

    if args.apply_access:
        logging.debug('apply_access starting ...')
        gitlab_usermgmt.gitlab_usermgmt.apply_access(
            args.gitlab_token,
            args.gitlab_url,
            "users_compared.json"
        )
        logging.debug('apply_access finished.')

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
