import sys
import json
import os
import logging

log = logging.getLogger(__name__)


def auth(token, url):
    """
    Authenticates against gitlab with your token
    :param token: The gitlab API token
    :param url: The gitlab URL, eg. https://gitlab.com
    :return:
    """
    import gitlab
    log.debug("auth: ...")
    gl = gitlab.Gitlab(url, private_token=token)
    gl.auth()
    return gl


def get_all_subgroups(gl, group_id):
    """
    Returns recursively all gitlab subgroup objects in a list of a source group
    :param gl: gitlab object
    :param group_id: source group id
    """
    log.debug("get_all_subgroups: ...")
    group_list = []
    group = gl.groups.get(group_id)
    group_list.append(group)
    subgroups = group.subgroups.list()
    for subgroup in subgroups:
        group_list = group_list + get_all_subgroups(gl, subgroup.id)
    return group_list


def get_all_projects(gl, subgroups):
    """
    Returns all gitlab project objects from a list of groups
    :param gl: gitlab object
    :param subgroups: list of gitlab group objects
    """
    log.debug("get_all_projects: ...")
    project_list = []
    for subgroup in subgroups:
        group_projects = subgroup.projects.list()
        for group_project in group_projects:
            project_list.append(gl.projects.get(group_project.id))
    return project_list


def update_user_dict_entry(user_dict_entry, username, state,
                           access_level, expires_at, group_id, web_url):
    """
    Updates a user entry with a access right

    :param user_dict_entry:
    :param username:
    :param state:
    :param access_level:
    :param expires_at:
    :param group_id:
    :param web_url:
    :return:
    """
    log.debug("update_user_dict_entry: ...")
    access_dict = dict(
        web_url=web_url,
        access_level=access_level,
        expires_at=expires_at
    )
    if user_dict_entry is None:
        user_dict_entry = dict(
            username=username,
            state=state,
            access=dict()
        )
    user_dict_entry['access'][group_id] = access_dict
    return user_dict_entry


def get_all_users(gl, subgroups, projects):
    """
    Returns a dict of users with their user access to groups and projects
    :param gl: gitlab object
    :param subgroups: list of gitlab group objects
    :param projects:  list of gitlab project objects
    :return: dict of users
    """
    log.debug("get_all_users: ...")
    user_dict = {}
    # get all users from groups first
    for subgroup in subgroups:
        group_members = subgroup.members.list(all=True)
        for group_member in group_members:
            user_dict[group_member.id] = update_user_dict_entry(
                user_dict.get(group_member.id),
                group_member.username,
                group_member.state,
                group_member.access_level,
                group_member.expires_at,
                subgroup.id,
                subgroup.web_url)
    # get all users from the projects
    for project in projects:
        project_members = project.members.list(all=True)
        for project_member in project_members:
            user_dict[project_member.id] = update_user_dict_entry(
                user_dict.get(project_member.id),
                project_member.username,
                project_member.state,
                project_member.access_level,
                project_member.expires_at,
                project.id,
                project.web_url)
    return user_dict


def write_dict_to_json(data, filename):
    """
    Writes a dict to a file

    :param data: a data dictionary
    :param filename: the filename it will save the json to
    :return:
    """
    log.debug("write_dict_to_json: ...")
    f = open(filename, "w")
    f.write(json.dumps(data))
    f.close()


def write_user_access_list(group_id, token, gitlab_url):
    """
    Logs in to Gitlab, queries all subgroups and projects,
    writes users.json file

    :param group_id: The source group id
    :param token: The gitlab API token
    :param gitlab_url: The gitlab server, eg. "https://gitlab.com"
    :return:
    """
    log.debug("write_user_access_list: ...")
    gl = auth(token, gitlab_url)
    subgroups = get_all_subgroups(gl, group_id)
    log.info("subgroups.len: " + str(subgroups.__len__()))
    projects = get_all_projects(gl, subgroups)
    log.info("projects.len: " + str(projects.__len__()))
    users = get_all_users(gl, subgroups, projects)
    log.info("users.len: " + str(users.__len__()))
    write_dict_to_json(users, "users.json")


def write_user_compared_list(users_file, access_def_file, users_compared_file):
    """
    Reads a list of user accesses, access definitions and compares them
    result will be written to a third file

    :param users_file: the current users with their access rights
    :param access_def_file: how the access should be
    :param users_compared_file: the user access file with the results
    :return:
    """
    log.debug("write_user_compared_list: ...")
    access_definition = read_json_to_dict(access_def_file)
    users = read_json_to_dict(users_file)
    users_compared = compare_user_access(users, access_definition)
    write_dict_to_json(users, users_compared_file)


def read_json_to_dict(filename):
    """
    Reads a json file and returns a dict of it

    :param filename: the input file
    :return:
    """
    log.debug("read_json_to_dict: filename: " + filename)
    f = open(filename, "r")
    return json.loads(f.read())


def compare_user_access(users, definition):
    """
    Compares the existing access rights with the rule set

    :param users: dict of users
    :param definition: dict of definitions
    :return:
    """
    log.debug("compare_user_access: ...")
    # iterate over all users an check their access
    for user_id, user_info in users.items():
        # initialize dict with access that will be removed
        user_info['to_be_removed'] = {}
        # initialize dict with access that will be changed
        user_info['to_be_changed'] = {}
        # initialize dict with access that will be granted
        user_info['to_be_granted'] = {}

        # iterate over all the current access entries of the user
        # if the user has access to a group/project but does not have the usergroup that allows that
        #  the group/project will be added to his to_be_removed dict
        # if the user has an access_level that does not matches the access_level defined in the
        #  access_definition usergroup, the defined group / project will be added to to_be_changed dict
        for user_access_id, user_access_info in user_info['access'].items():
            # if user is not active, remove all his access
            if user_info['state'] != "active":
                user_info['to_be_removed'][str(user_access_id)] = user_access_info
                continue
            current_access_level = user_access_info['access_level']
            defined_access_level = 0
            access_found = False
            # check if user has access but does not exist in definition list
            if user_id in definition['users']:
                # get the usergroups of the user
                definition_usergroups = definition['users'][str(user_id)]['usergroups']
                for usergroup in definition_usergroups:
                    # get the usergroup info
                    def_usergroup_info = definition['usergroups'][str(usergroup)]
                    # check if user_access_id is in usergroup
                    if str(user_access_id) in def_usergroup_info:
                        # check if access level matches of user if defined access level is unset
                        if defined_access_level == 0 and current_access_level == def_usergroup_info[str(user_access_id)]['access_level']:
                            defined_access_level = current_access_level
                        else:
                            # only update defined access level if its lower than the level in the current usergroup
                            if defined_access_level < def_usergroup_info[str(user_access_id)]['access_level']:
                                defined_access_level = def_usergroup_info[str(user_access_id)]['access_level']

            if defined_access_level == 0:
                user_info['to_be_removed'][str(user_access_id)] = user_access_info
            elif current_access_level != defined_access_level:
                user_info['to_be_changed'][str(user_access_id)] = {}
                user_info['to_be_changed'][str(user_access_id)]["web_url"] = user_access_info["web_url"]
                user_info['to_be_changed'][str(user_access_id)]["access_level"] = defined_access_level
                user_info['to_be_changed'][str(user_access_id)]["expires_at"] = user_access_info["expires_at"]

    # after comparing all the current access with the definition we know compare if all definitions
    #  mach the current access
    for def_user_id, def_user_info in definition['users'].items():
        for def_usergroup in def_user_info['usergroups']:
            for def_access_id, def_access_info in definition['usergroups'][def_usergroup].items():
                # know we have all groups and projects a user should have access to
                #  lets see if we can find the id in the users access list
                if str(def_user_id) in users:
                    user_info = users[str(def_user_id)]
                    if str(def_access_id) in user_info['access']:
                        pass
                    else:
                        # check if access_id is on grant list
                        if str(def_access_id) in user_info['to_be_granted']:
                            accesslevel_old = user_info['to_be_granted'][str(def_access_id)]['access_level']
                            if accesslevel_old < def_access_info['access_level']:
                                user_info['to_be_granted'][str(def_access_id)]['access_level'] = def_access_info[
                                    'access_level']
                        else:
                            user_info['to_be_granted'][str(def_access_id)] = def_access_info
                else:
                    # user id not found in users list, create a new user and add access to grant list
                    users[str(def_user_id)] = {}
                    users[str(def_user_id)]["username"] = def_user_info["username"]
                    users[str(def_user_id)]["access"] = {}
                    users[str(def_user_id)]["to_be_removed"] = {}
                    users[str(def_user_id)]["to_be_changed"] = {}
                    users[str(def_user_id)]["to_be_granted"] = {}
                    users[str(def_user_id)]['to_be_granted'][str(def_access_id)] = def_access_info
    return users


def update_api_input_members_entry(members_entry, entry_type, user_id, access_level, expire_date=None):
    """
    Updates an api_input_members entry

    :param members_entry: the existing entry
    :param entry_type: possible values create | change | remove
    :param user_id: the gitlab user id
    :param access_level: the gitlab access level that will be applied, nr. for entry_type: remove
    :param expire_date: set a expire date, syntax of the string: YEAR-MONTH-DAY, eg. 2054-06-25; default=None
    :return: the updated entry
    """
    log.debug("update_api_input_members_entry: ...")
    # if entry does not exist, initialise it
    if members_entry is None:
        members_entry = dict(
            {
                "members": {
                    "create": [],
                    "change": [],
                    "remove": []
                }
            }
        )
    members_entry['members'][entry_type] = members_entry['members'][entry_type] + [[user_id, access_level, expire_date]]
    return members_entry


def generate_api_input_from_users_compared(users_compared_dict):
    """
    Reads the content of the users_compared dictionary and writes it to a dict that
    can be directly applied to the gitlab API


    :param users_compared_dict: the dict with the contents of the users_compared.json
    :return: returns api_input dict
    """
    log.debug("generate_api_input_from_users_compared: ...")
    api_input = {'groups': {}, 'projects': {}}

    # iterate over all users
    for user_id, user_info in users_compared_dict.items():
        # iterate over to_be_removed list
        for access_id, access_info in user_info['to_be_removed'].items():
            # add entry to api_input remove dict
            if access_info['web_url'].__contains__('/groups/'):
                if access_id not in api_input['groups']:
                    api_input['groups'][str(access_id)] = None
                api_input['groups'][str(access_id)] = update_api_input_members_entry(
                    api_input['groups'][str(access_id)],
                    "remove",
                    user_id,
                    access_info['access_level'],
                    access_info['expires_at']
                )
            else:
                if access_id not in api_input['projects']:
                    api_input['projects'][str(access_id)] = None
                api_input['projects'][str(access_id)] = update_api_input_members_entry(
                    api_input['projects'][str(access_id)],
                    "remove",
                    user_id,
                    access_info['access_level'],
                    access_info['expires_at']
                )
        # iterate over to_be_changed list
        for access_id, access_info in user_info['to_be_changed'].items():
            # add entry to api_input change dict
            if access_info['web_url'].__contains__('/groups/'):
                if access_id not in api_input['groups']:
                    api_input['groups'][str(access_id)] = None
                api_input['groups'][str(access_id)] = update_api_input_members_entry(
                    api_input['groups'][str(access_id)],
                    "change",
                    user_id,
                    access_info['access_level'],
                    access_info['expires_at']
                )
            else:
                if access_id not in api_input['projects']:
                    api_input['projects'][str(access_id)] = None
                api_input['projects'][str(access_id)] = update_api_input_members_entry(
                    api_input['projects'][str(access_id)],
                    "change",
                    user_id,
                    access_info['access_level'],
                    access_info['expires_at']
                )
        # iterate over to_be_granted list
        for access_id, access_info in user_info['to_be_granted'].items():
            # add entry to api_input create dict
            if access_info['web_url'].__contains__('/groups/'):
                if access_id not in api_input['groups']:
                    api_input['groups'][str(access_id)] = None
                api_input['groups'][str(access_id)] = update_api_input_members_entry(
                    api_input['groups'][str(access_id)],
                    "create",
                    user_id,
                    access_info['access_level']
                )
            else:
                if access_id not in api_input['projects']:
                    api_input['projects'][str(access_id)] = None
                api_input['projects'][str(access_id)] = update_api_input_members_entry(
                    api_input['projects'][str(access_id)],
                    "create",
                    user_id,
                    access_info['access_level']
                )
    return api_input


def manage_gitlab_members(gl, accessobject, info):
    """
    Manages access for a group or a project based in info

    :param gl: gitlab object
    :param accessobject: a gitlab group or project object
    :param info: dict based with create / change / remove tuples
    :return:
    """
    import gitlab
    log.debug("manage_gitlab_members: accessobject: " + str(accessobject.id) + " info: " + str(info))
    for t in info["create"]:
        try:
            member = accessobject.members.create({
                'user_id': t[0],
                'access_level': t[1]
            })
            member.save()
        except gitlab.exceptions.GitlabCreateError as e:
            if e.response_code == 409:
                log.info("manage_gitlab_members: Member " + str(t[0]) + " already exists in " + str(accessobject.id))
            elif e.response_code == 400:
                log.info("manage_gitlab_members: Member " + str(t[0]) + " has higher inherited access level to " + str(
                    accessobject.id))
            else:
                raise e
    for t in info["change"]:
        member = accessobject.members.get(t[0])
        member.access_level = t[1]
        if t[2]:
            member.expires_at = t[2]
        member.save()
    for t in info["remove"]:
        try:
            accessobject.members.delete(t[0])
        except gitlab.exceptions.GitlabDeleteError as e:
            if e.response_code == 404:
                log.info("manage_gitlab_members: Member " + str(t[0]) + " already deleted in " + str(accessobject.id))
            else:
                raise e


def apply_access(token, gitlab_url, users_compared_file):
    """
    Logs in to Gitlab, queries all subgroups and projects,
    writes users.json file

    :param group_id: The source group id
    :param token: The gitlab API token
    :param gitlab_url: The gitlab server, eg. "https://gitlab.com"
    :param users_compared_file: the user access file with the results
    :return:
    """
    log.debug("apply_access: gitlab_url: " + gitlab_url + " users_compared_file: " + users_compared_file)
    gl = auth(token, gitlab_url)
    users_compared = read_json_to_dict(users_compared_file)
    api_input = generate_api_input_from_users_compared(users_compared)
    # write api_input to disk for compliance and debugging purposes
    write_dict_to_json(api_input, "api_input.json")
    # iterate over groups
    for group_id, group_info in api_input['groups'].items():
        log.debug("apply_access: group_id: " + group_id)
        group = gl.groups.get(group_id)
        manage_gitlab_members(gl, group, group_info['members'])
    # iterate over projects
    for project_id, project_info in api_input['projects'].items():
        log.debug("apply_access: project_id: " + project_id)
        project = gl.projects.get(project_id)
        manage_gitlab_members(gl, project, project_info['members'])


def manage_admins_only(users, expire_date=None):
    """
    Modifies users_compared dict.
    Entries on to_be_removed list with access < 40 will be removed.
    If these entries had no expire date, or an expire date that is higher then expire_date
    entry will be added to to_be_changed list with expire_date

    :param users: users_compared dict
    :param expire_date: string in the format of YEAR-MONTH-DAY, eg. 2021-03-28
    :return: updated users_compared dict
    """
    from datetime import datetime
    from datetime import date
    from dateutil.relativedelta import relativedelta
    if expire_date is None:
        expire_date = str(date.today() + relativedelta(years=1))
    log.debug("manage_admins_only: expire_date: " + str(expire_date))
    for user_id, user_info in users.items():
        # iterate over to_be_removed entries
        entries_that_will_be_deleted = []
        for entry_id, entry_info in user_info['to_be_removed'].items():
            # check if access level is lower than maintainer
            # if so, delete from remove list, add to change list if date must be modified
            if entry_info['access_level'] < 40:
                # check if expire_at must be changed
                expire_at = expire_date
                add_to_change_list = False
                # compare dates if an expire date exist
                if entry_info['expires_at'] is not None:
                    date_in = datetime.strptime(entry_info['expires_at'], "%Y-%m-%d")
                    date_cmp = datetime.strptime(expire_date, "%Y-%m-%d")
                    # check if entry info expire date is later than input parameter
                    if date_in > date_cmp:
                        add_to_change_list = True
                else:
                    add_to_change_list = True
                # add entry to change list
                if add_to_change_list:
                    user_info['to_be_changed'][str(entry_id)] = {}
                    user_info['to_be_changed'][str(entry_id)]["web_url"] = entry_info["web_url"]
                    user_info['to_be_changed'][str(entry_id)]["access_level"] = entry_info['access_level']
                    user_info['to_be_changed'][str(entry_id)]["expires_at"] = expire_at
                entries_that_will_be_deleted += [entry_id]
        # delete entry from remove list
        for entry_id in entries_that_will_be_deleted:
            user_info['to_be_removed'].pop(entry_id)
    return users


def remove_expire_dates(users):
    """
    Remove expire_at entries for managed access rights. Checks users_compared list of
    access and to_be_changed. When a expires_at entry is found that entry will be added to the
    to_be_changed list with expires_at=None.

    :param users: users_compared dict
    :return: updated users_compared dict
    """
    log.debug("remove_expire_dates: ...")
    for user_id, user_info in users.items():
        # iterate over to_be_changed first
        for entry_id, entry_info in user_info['to_be_changed'].items():
            # check if expires_at is set and set it to none
            if entry_info['expires_at'] is not None:
                entry_info['expires_at'] = None
        # iterate over access list
        for entry_id, entry_info in user_info['access'].items():
            # check if expires_at is set and set it to none
            if entry_info['expires_at'] is not None:
                # skip if entry will be removed anyway
                if entry_id in user_info['to_be_removed']:
                    continue
                # skip if entry is in to_be_changed list
                if entry_id in user_info['to_be_changed']:
                    continue
                # add entry to to_be_changed list with None expire date
                user_info['to_be_changed'][entry_id] = entry_info
                user_info['to_be_changed'][entry_id]['expires_at'] = None
    return users

def main():
    pass


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
